import java.util.*;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.sql.*;

public class TwoDimTreeManager
{
	private String databaseName = "";
	final static String SQLITE_URL = "jdbc:sqlite:";	
	
	public static void main(String[] args) throws SQLException, ClassNotFoundException
	{
		if (args.length != 2)
		{
			System.out.println("Please have 2 arguments!");
			return;
		}
		
		String databaseName = args[0];
		String instructionFile = args[1];
		
		BufferedReader reader;

		TwoDimTreeManager myTreeManager = new TwoDimTreeManager(databaseName);
		try {
			reader = new BufferedReader(new FileReader(instructionFile));
			String line = reader.readLine();
			
			// Read and process the input
			while (line != null) {
				System.out.println(line);
				String[] command = line.split(" ");
				String flag = command[0];
				
				if (flag.equals("c"))
				{
					myTreeManager.createTwoDimTree(command[1]);
				}
				else if (flag.equals("i"))
				{
					String tree_name = command[1];
					String label = command[2];
					float x_value = Float.valueOf(command[3]);
					float y_value = Float.valueOf(command[4]);
					
					TwoDimRecord record = new TwoDimRecord(-1, -1, label, x_value, y_value);
					
					myTreeManager.add(tree_name, record);
				}
				else if (flag.equals("l"))
				{
					String tree_name = command[1];
					float p_x = Float.valueOf(command[2]);
					float p_y = Float.valueOf(command[3]);
					TwoDimRecord record = myTreeManager.lookup(tree_name, p_x, p_y);
					
					if (record != null)
					{
						System.out.println(record.toString());
					}
					else
					{
						System.out.println("("+ p_x + ", " + p_y + ")");
					}	
				}
				else if (flag.equals("p"))
				{
					String tree_name = command[1];
					myTreeManager.printTree(tree_name);
				}
				else
				{
					System.out.println("Cannot process this command!");
				}
				System.out.println("");
				line = reader.readLine();
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
/**************************************************************************************************/
/*************************************** PUBLIC METHOD **********************************************/
	/**
	 * Constructor 
	 * @param databaseName
	 */
	public TwoDimTreeManager(String databaseName)
	{
		this.databaseName = databaseName;
		try {
			java.sql.Statement statement = null;
			Connection connection = null;
			
			connection = DriverManager.getConnection(SQLITE_URL + databaseName);		
			statement = connection.createStatement();
			
			// Create table TWO_DIM_TREE
			String two_dim_tree_String = "CREATE TABLE IF NOT EXISTS TWO_DIM_TREE( \n"
					+ "FILE_NAME VARCHAR(64), \n"
					+ "ROOT_ID INTEGER );";
			statement.executeUpdate(two_dim_tree_String);
			
			// Create table TWO_DIM_TREE_PART
			String two_dim_tree_part_string = "CREATE TABLE IF NOT EXISTS TWO_DIM_TREE_PART( \n"
					+ "NODE_ID INTEGER PRIMARY KEY AUTOINCREMENT, \n"
					+ "PARENT_ID INTEGER, \n"
					+ "IS_LEAF BOOLEAN, \n"
					+ "VALUE1 REAL, \n"
					+ "VALUE2 REAL);";
			statement.executeUpdate(two_dim_tree_part_string);
			
			
			// Create table TWO_DIM_AUX_NODE_DATA
			String two_dim_aux_node_data_string = "CREATE TABLE IF NOT EXISTS TWO_DIM_AUX_NODE_DATA( \n"
					+ "NODE_ID INTEGER, \n"
					+ "IS_X BOOLEAN, \n"
					+ "CHILD0 INTEGER, \n"
					+ "CHILD1 INTEGER, \n"
					+ "CHILD2 INTEGER, \n"
					+ "PRIMARY KEY(NODE_ID));";
			statement.executeUpdate(two_dim_aux_node_data_string);
			
			// Create table TWO_DIM_AUX_RECORD_DATA
			String two_dim_aux_record_data_string = "CREATE TABLE IF NOT EXISTS TWO_DIM_AUX_RECORD_DATA( \n"
					+ "NODE_ID INTEGER, \n"
					+ "LABEL CHAR(16), \n"
					+ "PRIMARY KEY(NODE_ID));";
			statement.executeUpdate(two_dim_aux_record_data_string);
			
			
			connection.close();
		} 
		catch (SQLException e) 
		{
			System.out.println("Error Connect to Database And Create Table!");
			
			e.printStackTrace();
		}
	}
	
	/** 
	 * This method create a tree with fileName
	 * @param fileName
	 * @return
	 */
	public boolean createTwoDimTree(String fileName)
	{
		
		try {
			Connection connection = null;
		
		
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);		
			
			PreparedStatement stmt = connection.prepareStatement(""
					+ "INSERT INTO TWO_DIM_TREE_PART(PARENT_ID, IS_LEAF)\n"
					+ "VALUES(?, ?);");
			stmt.setInt(1, -1);
			stmt.setBoolean(2, false);
			int count = stmt.executeUpdate();		
					
			
			// check if we can insert 
			if (count <= 0)
			{
				return false;
			}
			
			// Id is auto increase, so the largest Id means newest insert tree part
			stmt = connection.prepareStatement(
					"SELECT MAX(NODE_ID) AS ID\n"
					+ "FROM TWO_DIM_TREE_PART;");
			ResultSet rs = stmt.executeQuery();
			
			
			
			while (rs.next()) 
			{	
				// Insert into TWO_DIM_TREE the ID and file
				int rootID = rs.getInt("ID");
				PreparedStatement stmt2 = connection.prepareStatement(
						"INSERT INTO TWO_DIM_TREE(FILE_NAME, ROOT_ID) \n"
						+ "VALUES(?,?);");
				stmt2.setString(1, fileName);
				stmt2.setInt(2, rootID);
				count = stmt2.executeUpdate();
				
				if (count <= 0)
				{
					connection.close();
					return false;
				}
				
				
			}
			connection.close();
		
		} 
		catch (SQLException e) 
		{
			System.out.println("Error Create Two Dim Tree!");
			
			e.printStackTrace();
		}
		return true;
	}
	
	/**
	 * This method add a record a file name
	 * @param fileName
	 * @param record
	 * @return
	 */
	public boolean add(String fileName, TwoDimRecord record)
	{
		try {
			
			// variable set up
			int rootID = -1;
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);		
			
			
			//====================================================================
			// Get the rootID if tree exist
			PreparedStatement stmt = connection.prepareStatement(
					"SELECT ROOT_ID AS ID \n"
					+ "FROM TWO_DIM_TREE Tree \n"
					+ "WHERE Tree.FILE_NAME = ?;");	
			stmt.setString(1, fileName);
			ResultSet rs = stmt.executeQuery();
						
			if (!rs.next() ) // case tree do not exist
			{
			    System.out.println("Tree do not exist!");
			    return false;
			} 
			else // Case that tree exist
			{
				 rootID = rs.getInt("ID");			
			}
			connection.close();
			//====================================================================
			// Create A Leaf Node that hold data and get its ID
			createRecordNode(record);
			
			//===================================================================================
			// Insert that record Node to correct location	
			TwoDimNode nullParent = new TwoDimNode(-1, false, -1,-1,-1,-1,-1);
			
			TwoDimNode parentNode = searchParentLocation(record, rootID, nullParent);
    		
    		if (parentNode.id == -1) // Special case where root is empty
    		{
    			connection.close();
    			parentNode.id = rootID;
    			return addFirstRecord(record, parentNode);
    		}
    		else
    		{
    			connection.close();
    			return addNodeHelper(record, parentNode);
    			/** addNode(parentNode.id, newNodeID, true); */
    		}
    		
		} 
		catch (SQLException e) 
		{
			System.out.println("Error Create Two Dim Tree!");
			
			e.printStackTrace();
		}
		return false;

	}
	
	public TwoDimRecord lookup(String fileName, float x, float y)
	{
		TwoDimRecord record = new TwoDimRecord(-1, -1, "", x, y);
		TwoDimNode parentNode = new TwoDimNode(-1, false, -1,-1,-1,-1,-1);
		int rootID = -1;
					
		try {
			
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);			
			//====================================================================
			// Get the rootID if tree exist
			PreparedStatement stmt = connection.prepareStatement(
					"SELECT ROOT_ID AS ID \n"
					+ "FROM TWO_DIM_TREE Tree \n"
					+ "WHERE Tree.FILE_NAME = ?;");	
			stmt.setString(1, fileName);
			ResultSet rs = stmt.executeQuery();
						
			if (!rs.next() ) // case tree do not exist
			{
			    System.out.println("Tree do not exist!");
			    return null;
			} 
			else // Case that tree exist
			{
				 rootID = rs.getInt("ID");			
			}
			connection.close();
			parentNode = searchParentLocation(record, rootID, parentNode);
			
			if (parentNode.id == -1)
			{
				return null;
			}
			else
			{
				return getMatchingRecord(record, parentNode);
			}
		}
		catch (SQLException e) 
		{
			System.out.println("Error Look Up Record!");		
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * This method print the whole tree if it exist
	 * @param fileName
	 */
	@SuppressWarnings("resource")
	public void printTree(String fileName) 
	{
		LinkedList<Integer> queue = new LinkedList<Integer>();
		try {			
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);			
			//====================================================================
			// Get the rootID if tree exist
			PreparedStatement stmt = connection.prepareStatement(
					"SELECT ROOT_ID AS ID \n"
					+ "FROM TWO_DIM_TREE Tree \n"
					+ "WHERE Tree.FILE_NAME = ?;");	
			stmt.setString(1, fileName);
			ResultSet rs = stmt.executeQuery();
			
			if(!rs.next())
			{
				System.out.println("Tree does not exist!");
			}
			else
			{
				do
				{
					queue.add(rs.getInt("ID"));
				}
				while(rs.next());
			}
			
			int currentNode = 0;
			int parent = -1;
			boolean isLeaf = false;
			float value1 = -1;
			float value2 = -1;
			String value2String = "";
			int child0 = -1;
			int child1 = -1;
			int child2= -1;
			String label ="";
			boolean isX = false;
			TwoDimNode node = new TwoDimNode(-1, false, -1, -1, -1, -1, -1);
			TwoDimRecord record = new TwoDimRecord(-1, -1, "", -1, -1);
			
			
			System.out.println(fileName);
			while(queue.size() != 0)
			{
				currentNode = queue.pop();
				
				if (currentNode == -1)
				{
					continue;
				}
				
				stmt = connection.prepareStatement(
						"SELECT * \n"
						+ "FROM TWO_DIM_TREE_PART Tree_Part \n"
						+ "WHERE Tree_Part.NODE_ID = ?;");	
				stmt.setInt(1, currentNode);
				rs = stmt.executeQuery();
				
				while(rs.next())
				{
					parent = rs.getInt("PARENT_ID");
					isLeaf = rs.getBoolean("IS_LEAF");
					value1 = rs.getFloat("VALUE1");
					value2String = rs.getString("VALUE2");
					
					if (value2String == null)
					{
						value2 = 0f / 0f;
					}
					else
					{
						value2 = rs.getFloat("VALUE2");	
					}
					
				}
				
				if (!isLeaf)
				{
					stmt = connection.prepareStatement(
							"SELECT * \n"
							+ "FROM TWO_DIM_AUX_NODE_DATA \n"
							+ "WHERE NODE_ID = ?;");	
					stmt.setInt(1, currentNode);
					rs = stmt.executeQuery();
					
					while(rs.next())
					{
						isX = rs.getBoolean("IS_X");
						child0 = rs.getInt("CHILD0");
						child1 = rs.getInt("CHILD1");
						child2 = rs.getInt("CHILD2");
					}
					// Add all children to queue
					queue.add(child0);
					queue.add(child1);
					queue.add(child2);
					
					// Prepare node and print
					node.id = currentNode;
					node.isLeaf = isLeaf;
					node.childId0 = child0;
					node.childId1 = child1;
					node.childId2 = child2;
					node.value1 = value1;
					node.value2 = value2;
					node.parent_id = parent;	
					node.isX = isX;
					System.out.println(node.toString());
				}
				else
				{
					stmt = connection.prepareStatement(
							"SELECT * \n"
							+ "FROM TWO_DIM_AUX_RECORD_DATA \n"
							+ "WHERE NODE_ID = ?;");
					
					stmt.setInt(1, currentNode);
					rs = stmt.executeQuery();
					
					while(rs.next())
					{
						label = rs.getString("LABEL");
					}
					
					record.id = currentNode;
					record.isLeaf = isLeaf;
					record.value1 = value1;
					record.value2 = value2;
					record.parent_id = parent;	
					record.label = label;
					System.out.println(record.toString());
				}
			}
			
			connection.close();
			
		}
		catch (SQLException e) 
		{
			System.out.println("Error Print Two Dim Tree!");		
			e.printStackTrace();
		}

	}
	
/**************************************************************************************************/
/*************************************** HELPER METHOD **********************************************/
	
	/**
	 * This helper method decide what case of adding node we need
	 * @param record
	 * @param parentNode
	 * @return
	 */
	private boolean addNodeHelper(TwoDimRecord record, TwoDimNode parentNode)
	{
		if (parentNode.childId1 == -1) // case there is 2 slot
		{
			return addSecondRecord(record, parentNode);
		}
		else if(parentNode.childId2 == -1) // case there is 1 slot
		{
			return addThirdRecord(record, parentNode);
		}
		
		
		//========================================================================
		// Case there is zero slot , so we have to split	
		
		float recordValue = -1;
		//==========================================================================
		//Determine what node we need to split
		
		// Get either x or y of record based on parent slit key
		if (parentNode.isX)
		{
			recordValue = record.value1;
		}
		else
		{
			recordValue = record.value2;
		}
		

		//======================================
		if(recordValue <= parentNode.value1)
		{
			splitNode(record, parentNode, 0);
		}
		else if(recordValue <= parentNode.value2)
		{
			splitNode(record, parentNode, 1);
		}
		else
		{
			splitNode(record, parentNode, 2);
		}
			
	
		return true;
	}
	
	/**
	 * This helper method attach a first record to parent node
	 * @param record
	 * @param parentNode
	 * @return
	 */
	private boolean addFirstRecord(TwoDimRecord record, TwoDimNode parentNode)
	{
		try 
		{
			//==================================================================
			// Insert the parent to TWO_DIM_AUX_NODE_DATA
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);
			PreparedStatement stmt = connection.prepareStatement(
					"INSERT INTO TWO_DIM_AUX_NODE_DATA(NODE_ID, IS_X, CHILD0, CHILD1, CHILD2) \n"
					+ "VALUES(?,?,?,?,?);");
			stmt.setInt(1, parentNode.id);
			stmt.setBoolean(2, parentNode.isX);
			stmt.setInt(3, -1);
			stmt.setInt(4, -1);
			stmt.setInt(5, -1);
			int count = stmt.executeUpdate();
			
			if (count <= 0)
			{
				connection.close();
				return false;
			}
			
			float recordValue = record.value1;
			if(!parentNode.isX)
			{
				recordValue = record.value2;
			}
			
			
			// Update node path in TWO_DIM_AUX_NODE_DATA
			stmt = connection.prepareStatement(
	    			"UPDATE TWO_DIM_AUX_NODE_DATA \n"
	    			+ "SET CHILD0 = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, record.id);
	    	stmt.setInt(2, parentNode.id);
	    	stmt.executeUpdate();
	    	
	    	// Update parent_ID of new record node
	    	stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET PARENT_ID = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, parentNode.id);
	    	stmt.setInt(2, record.id);
	    	stmt.executeUpdate();
	    	
	    	// Update the query key of parent Node
	    	stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET VALUE1 = ? \n"
	    			+ "WHERE NODE_ID = ?");	 
	    	stmt.setString(1, String.valueOf(recordValue));
	    	stmt.setInt(2, parentNode.id);
	    	stmt.executeUpdate();
	    	
	    	parentNode.childId0 = record.id;
	    	parentNode.value1 = recordValue;
	    	
	    	connection.close();
	    
		}
		catch (SQLException e) 
		{
			System.out.println("Error Add First Record!");
			
			e.printStackTrace();
		}
		
		return true;
    	
	}

	/**
	 * This helper method attach a second record to parent node
	 * @param record
	 * @param parentNode
	 * @return
	 */
	private boolean addSecondRecord(TwoDimRecord record, TwoDimNode parentNode)
	{
		float child0_x = 0;
		float child0_y = 0;
		float child1_x = 0;
		float child1_y = 0;
		float parentValue1 = 0;
		int child0 = -1;
		int child1 = -1;
		boolean canUseX = false;
		boolean canUseY = false;
		boolean parentIsX = false;
	
		try 
		{	
			// Update the child1 Node parent to become the root
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);
			PreparedStatement stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET PARENT_ID = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, parentNode.id);
	    	stmt.setInt(2, record.id);
	    	stmt.executeUpdate();
			
			
			
	    	//==================================================================
	    	// get value of all current children
			stmt = connection.prepareStatement(
					"SELECT VALUE1, VALUE2 \n"
					+ "FROM TWO_DIM_TREE_PART \n"
					+ "WHERE NODE_ID = ?");
			stmt.setInt(1, parentNode.childId0);
			ResultSet rs = stmt.executeQuery();
			
			while(rs.next())
			{
				child0_x = rs.getFloat("VALUE1");
				child0_y = rs.getFloat("VALUE2");
			}
			
			
			child1_x = record.value1;
			child1_y = record.value2;
			
			
			//========================================================================================
			// Update the root node based on new insert value
			
			//  check if we can use x as a split key
			if (child0_x != child1_x)
			{
				canUseX= true;
			}
			
			// check if we can use y as a split key
			if (child0_y != child1_y)
			{
				canUseY = true;
			}
			
			//========================================================================================
			// organize child0. child1, value of parent
			// if current key is X and continue use X is possible, or cannot use Y => we have to use X as query key
			if ((parentNode.isX && canUseX) || (!canUseY))
			{
				parentIsX = true;
				// compare two child x value 
				if (child0_x < child1_x) // if new value is larger the order is the same
				{
					child0 = parentNode.childId0;
					child1 = record.id;
				}
				else // we swap two child otherwise
				{
					child1 = parentNode.childId0;
					child0 = record.id;
				}
				parentValue1 = Math.min(child0_x, child1_x);
			}
			
			else // we have to use Y as the split key
			{
				// compare two child y value 
				parentIsX = false;
				if (child0_y < child1_y) // if new value is larger the order is the same
				{
					child0 = parentNode.childId0;
					child1 = record.id;
				}
				else // we swap two child otherwise
				{
					child1 = parentNode.childId0;
					child0 = record.id;
				}	
				parentValue1 = Math.min(child0_y, child1_y);
			}
			
			
			//==============================================================================================
			// Update parent Node in TWO_DIM_AUX_NODE_DATA
			stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_AUX_NODE_DATA \n"
	    			+ "SET CHILD0 = ?, \n"
	    			+ "CHILD1 = ?, \n"
	    			+ "IS_X = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, child0);
	    	stmt.setInt(2, child1);
	    	stmt.setBoolean(3, parentIsX);
	    	stmt.setInt(4, parentNode.id);
	    	stmt.executeUpdate();
			
			// Update parent search Key value in TWO_DIM_TREE_PART
	    	stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET VALUE1 = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setString(1, String.valueOf(parentValue1));
	    	stmt.setInt(2, parentNode.id);
	    	stmt.executeUpdate();
				
	    	connection.close();
	    
		}
		catch (SQLException e) 
		{
			System.out.println("Error Add Second Record!");
			
			e.printStackTrace();
		}
		
		return true;
	}
	
	/**
	 * This helper method attach the third record to parent Node
	 * @param record
	 * @param parentNode
	 * @return
	 */
	private boolean addThirdRecord(TwoDimRecord record, TwoDimNode parentNode)
	{
		float child0_x = 0;
		float child0_y = 0;
		float child1_x = 0;
		float child1_y = 0;
		float child2_x = 0;
		float child2_y = 0;
		float parentValue1 = 0;
		float parentValue2 = 0;		
		int child0 = -1;
		int child1 = -1;
		int child2 = -1;		
		boolean canUseX = false;
		boolean canUseY = false;
		boolean parentIsX = false;
	
		try 
		{	
			// Update the new record Node parent to become the root
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);
			PreparedStatement stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET PARENT_ID = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, parentNode.id);
	    	stmt.setInt(2, record.id);
	    	stmt.executeUpdate();
			
	    	//==================================================================
	    	// get value of all current children
			stmt = connection.prepareStatement(
					"SELECT VALUE1, VALUE2 \n"
					+ "FROM TWO_DIM_TREE_PART \n"
					+ "WHERE NODE_ID = ?");
			stmt.setInt(1, parentNode.childId0);
			ResultSet rs = stmt.executeQuery();
			
			while(rs.next())
			{
				child0_x = rs.getFloat("VALUE1");
				child0_y = rs.getFloat("VALUE2");
			}
			
			stmt = connection.prepareStatement(
					"SELECT VALUE1, VALUE2 \n"
					+ "FROM TWO_DIM_TREE_PART \n"
					+ "WHERE NODE_ID = ?");
			stmt.setInt(1, parentNode.childId1);
			rs = stmt.executeQuery();
			
			while(rs.next())
			{
				child1_x = rs.getFloat("VALUE1");
				child1_y = rs.getFloat("VALUE2");
			}
			
		
			child2_x = record.value1;
			child2_y = record.value2;
			
			
			//========================================================================================
			// Update the root node based on new insert value
			
			// check if three x value is different
			if ((child0_x != child1_x) && (child1_x != child2_x) && (child2_x != child0_x))
			{
				canUseX= true;
			}
			
			// check if three x value is different
			if ((child0_y != child1_y) && (child1_y != child2_y) && (child2_y != child0_y))
			{
				canUseY= true;
			}
			
			//=================================================================================================
			// Organize 3 children and value1, value2 of parent
			// if current key is X and continue use X is possible, or cannot use Y => we have to use X as query key
			if ((parentNode.isX && canUseX) || (!canUseY))
			{
				parentIsX = true;
				Map<Float,Integer> myMap = new HashMap<Float,Integer>();
				myMap.put(child0_x, parentNode.childId0);
				myMap.put(child1_x, parentNode.childId1);
				myMap.put(child2_x, record.id);
				
				// Map 3 x value in order
				for ( Float key : myMap.keySet() ) {
				    if (key == Math.min(Math.min(child0_x, child1_x), child2_x))
				    {
				    	child0 = myMap.get(key);
				    	parentValue1 = key;
				    }
				    else if (key == Math.max(Math.max(child0_x, child1_x), child2_x))
				    {
				    	child2 = myMap.get(key);
				    }
				    else
				    {
				    	child1 = myMap.get(key);
				    	parentValue2 = key;
				    }
				}
			}
			
			else // we have to use Y as the split key
			{
				parentIsX = false;
				Map<Float,Integer> myMap = new HashMap<Float,Integer>();
				myMap.put(child0_y, parentNode.childId0);
				myMap.put(child1_y, parentNode.childId1);
				myMap.put(child2_y, record.id);
				
				// map 3 y values in order 
				for ( Float key : myMap.keySet() ) {
				    if (key == Math.min(Math.min(child0_y, child1_y), child2_y))
				    {
				    	child0 = myMap.get(key);
				    	parentValue1 = key;
				    }
				    else if (key == Math.max(Math.max(child0_y, child1_y), child2_y))
				    {
				    	child2 = myMap.get(key);
				    }
				    else
				    {
				    	child1 = myMap.get(key);
				    	parentValue2 = key;
				    }
				}
			}	
				
			//==============================================================================================
			// Update parent Node in TWO_DIM_AUX_NODE_DATA
			stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_AUX_NODE_DATA \n"
	    			+ "SET CHILD0 = ?, \n"
	    			+ "CHILD1 = ?, \n"
	    			+ "CHILD2 = ?, \n"
	    			+ "IS_X = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, child0);
	    	stmt.setInt(2, child1);
	    	stmt.setInt(3, child2);
	    	stmt.setBoolean(4, parentIsX);
	    	stmt.setInt(5, parentNode.id);
	    	stmt.executeUpdate();
			
			// Update parent Key in TWO_DIM_TREE_PART
	    	stmt = connection.prepareStatement(
	    			"UPDATE  TWO_DIM_TREE_PART \n"
	    			+ "SET VALUE1 = ?, \n"
	    			+ "VALUE2 = ? \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setString(1, String.valueOf(parentValue1));
	    	stmt.setString(2, String.valueOf(parentValue2));
	    	stmt.setInt(3, parentNode.id);
	    	stmt.executeUpdate();
				
	    	connection.close();
	    
		}
		catch (SQLException e) 
		{
			System.out.println("Error Add Third Record!");
			
			e.printStackTrace();
		}
		
		return true;
	}

	/**
	 * This method split the children of a parent and also attach the new record to it
	 * @param record
	 * @param parentNode
	 * @param splitChildNumber
	 */
	private void splitNode(TwoDimRecord record, TwoDimNode parentNode, int splitChildNumber) 
	{
		TwoDimNode newInnerNode = new TwoDimNode(-1, !parentNode.isX, -1, -1, -1,-1,-1);
		TwoDimRecord previousLeaveRecord = new TwoDimRecord(-1,-1,"",-1,-1);
		try 
		{
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);
			
			// Create new InnerNode
			PreparedStatement stmt = connection.prepareStatement(
					"INSERT INTO TWO_DIM_TREE_PART(PARENT_ID, IS_LEAF) \n"
					+ "VALUES(?, ?);");
			stmt.setInt(1, parentNode.id);
			stmt.setBoolean(2, false);
			stmt.executeUpdate();
			
			
			// get its ID
			stmt = connection.prepareStatement(
					"SELECT MAX(NODE_ID) AS ID \n"
					+ "FROM TWO_DIM_TREE_PART;");
			ResultSet rs = stmt.executeQuery();
			
			while(rs.next())
			{
				newInnerNode.id = rs.getInt("ID");
			}
			
			// Update parentNode child pointer point to new innerNode
			if (splitChildNumber == 0)
			{
				previousLeaveRecord.id = parentNode.childId0;
		    	
			}
			else if (splitChildNumber == 1)
			{
				previousLeaveRecord.id = parentNode.childId1;
			}
			else
			{
				previousLeaveRecord.id = parentNode.childId2;
			}
			
			stmt = connection.prepareStatement(
	    			"UPDATE TWO_DIM_AUX_NODE_DATA \n"
	    			+ "SET CHILD" + String.valueOf(splitChildNumber) + " = ? \n"
	    			+ "WHERE NODE_ID = ?");
			
	    	stmt.setInt(1, newInnerNode.id);
	    	stmt.setInt(2, parentNode.id);
	    	stmt.executeUpdate();
			
	    	stmt = connection.prepareStatement(
	    			"SELECT * \n"
	    			+ "FROM TWO_DIM_TREE_PART \n"
	    			+ "WHERE NODE_ID = ?");
	    	stmt.setInt(1, previousLeaveRecord.id);
	    	rs = stmt.executeQuery();
	    	
	    	while(rs.next())
	    	{
	    		previousLeaveRecord.value1 = rs.getFloat("VALUE1");
	    		previousLeaveRecord.value2 = rs.getFloat("VALUE2"); 		
	    	}
			addFirstRecord(previousLeaveRecord, newInnerNode);
			addSecondRecord(record, newInnerNode);
	    	
		}
		catch (SQLException e) 
		{
			System.out.println("Error split Node!");
			
			e.printStackTrace();
		}
			
	}
	
	/**
	 * This Helper method create a leave node that hold data
	 * @param x
	 * @param y
	 * @param label
	 * @return the Node_ID of that new created noded
	 */
	private void createRecordNode(TwoDimRecord record)
	{
		int newNodeID = -1;
		try {
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);	
			
			//=================================================================
			// Create The Leave Node hold coordinate in TWO_DIM_TREE_PART
			PreparedStatement stmt = connection.prepareStatement(
					"INSERT INTO TWO_DIM_TREE_PART(PARENT_ID, IS_LEAF, VALUE1, VALUE2) \n"
					+ "VALUES(?, ?, ?, ?);");
			stmt.setInt(1, -1);
			stmt.setBoolean(2, true);
			stmt.setString(3, String.valueOf(record.value1));
			stmt.setString(4, String.valueOf(record.value2));
			stmt.executeUpdate();
	
			//=================================================================
			// Get ID of new record Node
			stmt = connection.prepareStatement(
					"SELECT MAX(NODE_ID) AS ID\n"
					+ "FROM TWO_DIM_TREE_PART;");
			ResultSet rs = stmt.executeQuery();
			
			
			while (rs.next()) 
			{
				newNodeID = rs.getInt("ID");						
			}
			
			//=================================================================
			// Record the Data  "label" corresponding to that leaf node in TWO_DIM_AUX_RECORD_DATA
			stmt = connection.prepareStatement(
					"INSERT INTO TWO_DIM_AUX_RECORD_DATA(NODE_ID, LABEL) \n"
					+ "VALUES(?,?);");
			
			stmt.setInt(1, newNodeID);
			stmt.setString(2, record.label);
			stmt.executeUpdate();
			
			//=================================================================
			// Update recordID as well
			record.id = newNodeID;
			
			connection.close();
		}
		catch (SQLException e) 
		{
			System.out.println("Error Create Record!");
			
			e.printStackTrace();
		}

	}

	/**
	 * This method search for the parent Node that the Node will attach to
	 * @param record
	 * @param currentNodeID
	 * @param parentNode
	 * @return
	 */
	private TwoDimNode searchParentLocation(TwoDimRecord record, int currentNodeID, TwoDimNode parentNode)
	{
		// Information of current node
		int child0 = -1;
		int child1 = -1;
		int child2 = -1;
		boolean isX = true;
		
		// two value of current node
		float value1 = -1;
		float value2 = -1;
		
		// value that we need to the record to compare
		float recordValue = -1;
			
		try 
		{
			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);
			
			// Get the all child of current parents
			PreparedStatement stmt = connection.prepareStatement(
					"SELECT * \n"
					+ "FROM  TWO_DIM_AUX_NODE_DATA Tree_Node \n"
					+ "WHERE Tree_Node.NODE_ID = ?;");
		
			stmt.setInt(1, currentNodeID);
			ResultSet rs = stmt.executeQuery();
			
		
			if(!rs.next()) // we return the parent Node if at the leaf node
			{
				connection.close();
				return parentNode;
			}
			else
			{
				do
				{	
				child0 = rs.getInt("CHILD0");
				child1 = rs.getInt("CHILD1");
				child2 = rs.getInt("CHILD2");
				isX = rs.getBoolean("IS_X");
				}
				while(rs.next());
			}
			
			// Set the record value based on isX
			if (isX)
			{
				recordValue = record.value1;
			}
			else
			{
				recordValue = record.value2;
			}
			
				
			// We come to next tree level based on the value of record
			stmt = connection.prepareStatement(
					"SELECT * \n"
					+ "FROM  TWO_DIM_TREE_PART Tree_Part \n"
					+ "WHERE Tree_Part.NODE_ID = ?;");
			stmt.setInt(1, currentNodeID);
			rs = stmt.executeQuery();
			
			while (rs.next())
			{
				value1 = rs.getFloat("VALUE1");
				value2 = rs.getFloat("VALUE2");
			}

			//=====================================================================================
			// Promote this currentNode to parent Node
			parentNode.childId0 = child0;
			parentNode.childId1 = child1;
			parentNode.childId2 = child2;
			parentNode.isX = isX;
			parentNode.value1 = value1;
			parentNode.value2 = value2;
			parentNode.id = currentNodeID;
		
			if (recordValue <= value1)
			{
				connection.close();
			
				return searchParentLocation(record, child0, parentNode);
			}
			else if ((recordValue > value1 && recordValue <= value2) ||
					(child1 == -1))
			{
				connection.close();
			
				return searchParentLocation(record, child1, parentNode);
			}
			else
			{
				connection.close();

				return searchParentLocation(record, child2, parentNode);
			}
			
		} 
		catch (SQLException e) 
		{
			System.out.println("Error Search Parent Record!");
			
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * This method get the child that match with coordinate of the record
	 * @param record
	 * @param parentNode
	 * @return
	 */
	private TwoDimRecord getMatchingRecord(TwoDimRecord record, TwoDimNode parentNode)
	{
		try {

			Connection connection = null;
			connection = DriverManager.getConnection(SQLITE_URL + this.databaseName);		
			PreparedStatement stmt = connection.prepareStatement(
					"SELECT * \n"
					+"FROM TWO_DIM_TREE_PART Tree_Part, TWO_DIM_AUX_RECORD_DATA Tree_Record \n"
					+"WHERE Tree_Part.NODE_ID = ? AND \n"
					+"Tree_Record.NODE_ID = ?;");
			
			// Case child 0 is correct
			stmt.setInt(1, parentNode.childId0);
			stmt.setInt(2, parentNode.childId0);
			ResultSet rs = stmt.executeQuery();

			while(rs.next())
			{
				if ((rs.getFloat("VALUE1") == record.value1) &&
					(rs.getFloat("VALUE2") == record.value2))
				{
					record.id = rs.getInt("NODE_ID");
					record.label = rs.getString("LABEL");
					record.parent_id = rs.getInt("PARENT_ID");
					return record;
				}
			}
			
			// Case child 1 is correct
			stmt.setInt(1, parentNode.childId1);
			stmt.setInt(2, parentNode.childId1);
			rs = stmt.executeQuery();
			while(rs.next())
			{
				if ((rs.getFloat("VALUE1") == record.value1) &&
					(rs.getFloat("VALUE2") == record.value2))
				{
					record.id = rs.getInt("NODE_ID");
					record.label = rs.getString("LABEL");
					record.parent_id = rs.getInt("PARENT_ID");
					return record;
				}
			}
			
			// Case child 2 is correct
			stmt.setInt(1, parentNode.childId2);
			stmt.setInt(2, parentNode.childId2);
			rs = stmt.executeQuery();
			while(rs.next())
			{
				if ((rs.getFloat("VALUE1") == record.value1) &&
					(rs.getFloat("VALUE2") == record.value2))
				{
					record.id = rs.getInt("NODE_ID");
					record.label = rs.getString("LABEL");
					record.parent_id = rs.getInt("PARENT_ID");
					return record;
				}
			}
						
		} 
		catch (SQLException e) 
		{
			System.out.println("Error Get Matching Record!");
			
			e.printStackTrace();
		}
		
		return null;
	}
}